// vincia08-root is a part of the VINCIA plugin to PYTHIA 8
// This program runs several instances of PYTHIA 8 in parallel, one of which
// is standard PYTHIA 8, the others use different VINCIA settings. 
// All results are plotted simultaneously using the VinciaRoot interface
// Some observables corresponding to compressed hierarchies are computed.
// Copyright (c) 2017 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia/Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot/VinciaRoot.h"
// FASTJET
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Auxiliary routine to compute angles between jets
double jetAngle(fastjet::PseudoJet jet1, fastjet::PseudoJet jet2) {
    Vec4 p1(jet1.px(),jet1.py(),jet1.pz(),jet1.e());
    Vec4 p2(jet2.px(),jet2.py(),jet2.pz(),jet2.e());
    return theta(p1,p2);
}

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects
  Pythia pythiaA; 
  Pythia pythiaB; 
  Pythia pythiaC; 
  Pythia pythiaD; 
  VinciaPlugin vinciaA(&pythiaA);
  VinciaPlugin vinciaB(&pythiaB);
  VinciaPlugin vinciaC(&pythiaC);
  VinciaPlugin vinciaD(&pythiaD);
  // Read user settings from files
  pythiaA.readFile("vincia08a.cmnd");
  pythiaB.readFile("vincia08b.cmnd");
  pythiaC.readFile("vincia08c.cmnd");
  pythiaD.readFile("vincia08d.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Pointers to A and B settings
  Settings* settingsPtrA = vinciaA.settingsPtr;     
  Settings* settingsPtrB = vinciaB.settingsPtr;     
  Settings* settingsPtrC = vinciaC.settingsPtr;     
  Settings* settingsPtrD = vinciaD.settingsPtr;     
  bool   vinciaOnA       = settingsPtrA->flag("Vincia");
  bool   vinciaOnB       = settingsPtrB->flag("Vincia");
  bool   vinciaOnC       = settingsPtrC->flag("Vincia");
  bool   vinciaOnD       = settingsPtrD->flag("Vincia");

  // Extract settings to be used in the main program.  
  double eCMa      = settingsPtrA->parm("Beams:eCM");
  double eCMb      = settingsPtrB->parm("Beams:eCM");
  double eCMc      = settingsPtrC->parm("Beams:eCM");
  double eCMd      = settingsPtrD->parm("Beams:eCM");
  int    nEvent    = settingsPtrA->mode("Main:numberOfEvents");
  int    verbose   = settingsPtrB->mode("Vincia:verbose");

  //************************************************************************
  // VinciaRoot
  //************************************************************************

  // Create VinciaRoot object 
  VinciaRoot vroot(settingsPtrA); 
  vroot.init();
  
  // Change default to logarithmic Y axis
  vroot.setLogY();

  // Tell VinciaRoot how verbose you want it to be
  vroot.setVerbose("all",verbose);

  // Tell it the name of the MC we are running (for plot legends)
  string pyVer    = num2str(settingsPtrB->parm("Pythia:versionNumber"),5);  
  string vcVer    = num2str(settingsPtrB->parm("Vincia:versionNumber"),5);  

  // Tell it the name of the MC we are running (for plot legends)
  vector<string> legend;
  legend.push_back("A");
  legend.push_back("B");
  legend.push_back("C");
  legend.push_back("D");
  legend.push_back("E");
  legend.push_back("F");
  legend.push_back("G");
  legend.push_back("H");
  vroot.setLegend("all",legend);

  string MCnames = "Vincia "+vcVer;
  MCnames += " + MadGraph 4.426";
  MCnames += " + Pythia "+pyVer;
  vroot.setMCnames("all",MCnames);

  //************************************************************************
  // Initialize PYTHIA (automatically initializes VINCIA) 
  //************************************************************************

  // Initialize
  vinciaA.init();
  vinciaB.init();
  vinciaC.init();
  vinciaD.init();

  //***************************************************************************
  // FASTJET
  //***************************************************************************

  // Fastjet analysis - select algorithm and parameters
  fastjet::Strategy               strategy = fastjet::Best;
  fastjet::RecombinationScheme    recombScheme = fastjet::E_scheme;
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::ee_kt_algorithm,
                                      recombScheme, strategy);
  // Fastjet input
  std::vector <fastjet::PseudoJet> fjInputs;

  //***************************************************************************
  // ROOT
  //***************************************************************************

  //****************
  // Book Histograms
  double xLo, xHi;
  int nBins;
  int sizX = 480;
  int sizY = 550;

  // Jet resolution ratios:
  // Want to probe perturbative physics above scale for B decay, so 
  // restrict region of interest to 
  //   all yij > 0.005 ~=> ln(yij) > -5.3   ~=>   kT > 6.5 GeV
  // Also require y(ij-n) > 0.5 y(23)   => compressed hierarchies
  // Implies y(23) > 0.01   =>   ln(y23) > -4.6

  // Same Side LnY24
  nBins = 7;
  xLo = -1.1;
  xHi =  0.3;
  double wLnY24SS = nBins/(xHi-xLo);
  vector<TH1D*> VLnY24SS;
  VLnY24SS.push_back(new TH1D("LnY24SS","SameSide Ln(y34/y23)",nBins, xLo, xHi));
  VLnY24SS.push_back(new TH1D("LnY24SSb","SameSide Ln(y34/y23)",nBins, xLo, xHi));
  VLnY24SS.push_back(new TH1D("LnY24SSc","SameSide Ln(y34/y23)",nBins, xLo, xHi));
  VLnY24SS.push_back(new TH1D("LnY24SSd","SameSide Ln(y34/y23)",nBins, xLo, xHi));
  VLnY24SS[0]->SetMaximum(2.0); 
  VLnY24SS[0]->SetMinimum(0.005); 
  VLnY24SS[0]->GetXaxis()->SetTitle("#xi_{24}");
  VLnY24SS[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("LnY24SS",sizX,sizY);
  for (unsigned int iHis=0; iHis < VLnY24SS.size(); ++iHis) {
    vroot.addTH1DPtr("LnY24SS",VLnY24SS[iHis]);  
  }
  vroot.setNormalize("LnY24SS",true);
  
  // Same Side LnY25
  nBins =  6;
  xLo = -2.75;
  xHi =  0.25;
  double wLnY25SS = nBins/(xHi-xLo);
  vector<TH1D*> VLnY25SS;
  VLnY25SS.push_back(new TH1D("LnY25SS","SameSide Ln(y45/y23)",nBins, xLo, xHi));
  VLnY25SS.push_back(new TH1D("LnY25SSb","SameSide Ln(y45/y23)",nBins, xLo, xHi));
  VLnY25SS.push_back(new TH1D("LnY25SSc","SameSide Ln(y45/y23)",nBins, xLo, xHi));
  VLnY25SS.push_back(new TH1D("LnY25SSd","SameSide Ln(y45/y23)",nBins, xLo, xHi));
  VLnY25SS[0]->SetMaximum(2.0); 
  VLnY25SS[0]->SetMinimum(0.005); 
  VLnY25SS[0]->GetXaxis()->SetTitle("#xi_{25}");
  VLnY25SS[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("LnY25SS",sizX,sizY);
  for (unsigned int iHis=0; iHis < VLnY25SS.size(); ++iHis) {
    vroot.addTH1DPtr("LnY25SS",VLnY25SS[iHis]);  
  }
  vroot.setNormalize("LnY25SS",true);

  // Jet mass ratio
  nBins =   20;
  xLo   =  0.0;
  xHi   =  1.0;
  double wDM = nBins/(xHi-xLo);
  vector<TH1D*> VDM;
  VDM.push_back(new TH1D("DM","M^{2}_{L}/M^{2}_{H} with cuts",nBins, xLo, xHi));
  VDM.push_back(new TH1D("DMb","M^{2}_{L}/M^{2}_{H} with cuts",nBins, xLo, xHi));
  VDM.push_back(new TH1D("DMc","M^{2}_{L}/M^{2}_{H} with cuts",nBins, xLo, xHi));
  VDM.push_back(new TH1D("DMd","M^{2}_{L}/M^{2}_{H} with cuts",nBins, xLo, xHi));
  VDM[0]->SetMaximum(2.5); 
  VDM[0]->SetMinimum(-0.025*VDM[0]->GetMaximum()); 
  VDM[0]->GetXaxis()->SetTitle("M^{2}_{L}/M^{2}_{H}");
  VDM[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("DM",sizX,sizY);
  for (unsigned int iHis=0; iHis < VDM.size(); ++iHis) {
    vroot.addTH1DPtr("DM",VDM[iHis]);  
  }
  vroot.setNormalize("DM",true);

  // Jet mass ratio of ratios 
  // Left side of DM histogram divided by right side, defined as:
  //   N(mL2/mH2) < xL
  //   ---------------    as function of xL
  //   N(mL2/mH2) > xL   
  nBins =   7;
  xLo   =  0.175;
  xHi   =  0.525;
  vector<TH1D*> VDMratio;
  VDMratio.push_back(new TH1D("DMratio","N_{LM}/N_{MM}",nBins, xLo, xHi));
  VDMratio.push_back(new TH1D("DMratiob","N_{LM}/N_{MM}",nBins, xLo, xHi));
  VDMratio.push_back(new TH1D("DMratioc","N_{LM}/N_{MM}",nBins, xLo, xHi));
  VDMratio.push_back(new TH1D("DMratiod","N_{LM}/N_{MM}",nBins, xLo, xHi));
  VDMratio[0]->SetMaximum(1.21); 
  VDMratio[0]->SetMinimum(0.11); 
  VDMratio[0]->GetXaxis()->SetTitle("x_{L}");
  VDMratio[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("DMratio",sizX,sizY);
  //vroot.setLogY("DMratio");
  for (unsigned int iHis=0; iHis < VDMratio.size(); ++iHis) {
    vroot.addTH1DPtr("DMratio",VDMratio[iHis]);  
  }
  vroot.setNormalize("DMratio",false);
  vroot.setLogY("DMratio",false);

  // SameSide fractions 
  nBins =   3;
  xLo   =  1.5;
  xHi   =  4.5;
  vector<TH1D*> VSS;
  VSS.push_back(new TH1D("SS","SameSide Clustering Fractions",nBins, xLo, xHi));
  VSS.push_back(new TH1D("SSb","SameSide Clustering Fractions",nBins, xLo, xHi));
  VSS.push_back(new TH1D("SSc","SameSide Clustering Fractions",nBins, xLo, xHi));
  VSS.push_back(new TH1D("SSd","SameSide Clustering Fractions",nBins, xLo, xHi));
  VSS[0]->SetMaximum(1.0); 
  VSS[0]->SetMinimum(0.0005); 
  VSS[0]->GetXaxis()->SetTitle("N(SameSide)");
  VSS[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("SS Fractions",sizX,sizY);
  for (unsigned int iHis=0; iHis < VSS.size(); ++iHis) {
    vroot.addTH1DPtr("SS Fractions",VSS[iHis]);  
  }
  vroot.setNormalize("SS Fractions",true);

  // x4 = Energy fraction of fourth jet (control)
  nBins = 30;
  xLo =  0.0;
  xHi =  0.6;
  double wx4 = nBins/(xHi-xLo);
  vector<TH1D*> Vx4;
  Vx4.push_back(new TH1D("x4","x_{4}",nBins, xLo, xHi));
  Vx4.push_back(new TH1D("x4b","x_{4}",nBins, xLo, xHi));
  Vx4.push_back(new TH1D("x4c","x_{4}",nBins, xLo, xHi));
  Vx4.push_back(new TH1D("x4d","x_{4}",nBins, xLo, xHi));
  Vx4[0]->SetMaximum(6.01); 
  Vx4[0]->SetMinimum(-0.025*Vx4[0]->GetMaximum()); 
  Vx4[0]->GetXaxis()->SetTitle("x_{4}");
  Vx4[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("x4",sizX,sizY);
  for (unsigned int iHis=0; iHis < Vx4.size(); ++iHis) {
    vroot.addTH1DPtr("x4",Vx4[iHis]);  
  }
  vroot.setNormalize("x4",true);

  // the14 
  nBins =   9;
  xLo   =   0.0;
  xHi   = 180.0;
  double wthe14 = nBins/(xHi-xLo);
  vector<TH1D*> Vthe14;
  Vthe14.push_back(new TH1D("the14","#theta_{14}",nBins, xLo, xHi));
  Vthe14.push_back(new TH1D("the14b","#theta_{14}",nBins, xLo, xHi));
  Vthe14.push_back(new TH1D("the14c","#theta_{14}",nBins, xLo, xHi));
  Vthe14.push_back(new TH1D("the14d","#theta_{14}",nBins, xLo, xHi));
  Vthe14[0]->SetMaximum(0.017); 
  Vthe14[0]->SetMinimum(-0.025*Vthe14[0]->GetMaximum()); 
  Vthe14[0]->GetXaxis()->SetTitle("#theta");
  Vthe14[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("the14",sizX,sizY);
  for (unsigned int iHis=0; iHis < Vthe14.size(); ++iHis) {
    vroot.addTH1DPtr("the14",Vthe14[iHis]);  
  }
  vroot.setLegend("the14",legend,1);
  vroot.setNormalize("the14",true);

  // cos14 
  nBins =   20;
  xLo   =  -1.0;
  xHi   =   1.0;
  double wcos14 = nBins/(xHi-xLo);
  vector<TH1D*> Vcos14;
  Vcos14.push_back(new TH1D("cos14","cos(#theta_{14})",nBins, xLo, xHi));
  Vcos14.push_back(new TH1D("cos14b","cos(#theta_{14})",nBins, xLo, xHi));
  Vcos14.push_back(new TH1D("cos14c","cos(#theta_{14})",nBins, xLo, xHi));
  Vcos14.push_back(new TH1D("cos14d","cos(#theta_{14})",nBins, xLo, xHi));
  Vcos14[0]->SetMaximum(1.4); 
  Vcos14[0]->SetMinimum(-0.025*Vcos14[0]->GetMaximum()); 
  Vcos14[0]->GetXaxis()->SetTitle("cos(#theta)");
  Vcos14[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("cos14",sizX,sizY);
  for (unsigned int iHis=0; iHis < Vcos14.size(); ++iHis) {
    vroot.addTH1DPtr("cos14",Vcos14[iHis]);  
  }
  vroot.setNormalize("cos14",true);

  // Cos14 Ratios
  nBins =   5;
  xLo   =  0.5;
  xHi   =  5.5;
  vector<TH1D*> VNF;
  VNF.push_back(new TH1D("NF","Binned #theta_{14} Ratios",nBins, xLo, xHi));
  VNF.push_back(new TH1D("NFb","Binned #theta_{14} Ratios",nBins, xLo, xHi));
  VNF.push_back(new TH1D("NFc","Binned #theta_{14} Ratios",nBins, xLo, xHi));
  VNF.push_back(new TH1D("NFd","Binned #theta_{14} Ratios",nBins, xLo, xHi));
  VNF[0]->SetMaximum(1.25); 
  VNF[0]->SetMinimum(0.25); 
  VNF[0]->GetXaxis()->SetTitle("");
  VNF[0]->GetXaxis()->SetBinLabel(1,"A/T");
  VNF[0]->GetXaxis()->SetBinLabel(2,"Ai/Ti");
  VNF[0]->GetXaxis()->SetBinLabel(3,"C/F");
  VNF[0]->GetXaxis()->SetBinLabel(4,"Ci/F");
  VNF[0]->GetXaxis()->SetBinLabel(5,"Ci/Fi");
  VNF[0]->GetYaxis()->SetTitle("Ratio");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("Theta14ratios",sizX,sizY);
  for (unsigned int iHis=0; iHis < VNF.size(); ++iHis) {
    vroot.addTH1DPtr("Theta14ratios",VNF[iHis]);  
  }
  vroot.setNormalize("Theta14ratios",false);
  vroot.setLogY("Theta14ratios",false);

  // alphaStar
  nBins =   20;
  xLo   =  -0.1;
  xHi   =   0.9;
  double walphaStar = nBins/(xHi-xLo);
  vector<TH1D*> ValphaStar;
  ValphaStar.push_back(new TH1D("alphaStar","#alpha^{*}",nBins, xLo, xHi));
  ValphaStar.push_back(new TH1D("alphaStarb","#alpha^{*}",nBins, xLo, xHi));
  ValphaStar.push_back(new TH1D("alphaStarc","#alpha^{*}",nBins, xLo, xHi));
  ValphaStar.push_back(new TH1D("alphaStard","#alpha^{*}",nBins, xLo, xHi));
  ValphaStar[0]->SetMaximum(2.8); 
  ValphaStar[0]->SetMinimum(-0.025*ValphaStar[0]->GetMaximum()); 
  ValphaStar[0]->GetXaxis()->SetTitle("#alpha^{*}/#pi");
  ValphaStar[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("alphaStar",sizX,sizY);
  for (unsigned int iHis=0; iHis < ValphaStar.size(); ++iHis) {
    vroot.addTH1DPtr("alphaStar",ValphaStar[iHis]);  
  }
  vroot.setLegend("alphaStar",legend,-1);
  vroot.setNormalize("alphaStar",true);

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settingsPtrA->mode("Random:seed");
  vinciaA.rndmPtr->init(initseed);
  vinciaB.rndmPtr->init(initseed);
  vinciaC.rndmPtr->init(initseed);
  vinciaD.rndmPtr->init(initseed);

  // FastJet needs bool to tell whether this event is the first one
  bool firstEvent = true; 

  // Begin event loop
  double sumNa(0.), sumNb(0.), sumNc(0.), sumNd(0.);
  vector<double> sumZuds, sumZcc, sumZbb;
  vector<double> nLM2, nLM25, nLM3, nLM35, nLM4, nLM45, nLM5;
  vector<double> nMM2, nMM25, nMM3, nMM35, nMM4, nMM45, nMM5;
  vector<double> nTwd, nTrns, nAway;
  vector<double> nTwdI, nTrnsI, nAwayI;
  for (int iGen = 0; iGen < 4; ++iGen) {
    sumZuds.push_back(0.);
    sumZbb.push_back(0.);
    sumZcc.push_back(0.);
    nLM2.push_back(0.);
    nLM25.push_back(0.);
    nLM3.push_back(0.);
    nLM35.push_back(0.);
    nLM4.push_back(0.);
    nLM45.push_back(0.);
    nLM5.push_back(0.);
    nMM2.push_back(0.);
    nMM25.push_back(0.);
    nMM3.push_back(0.);
    nMM35.push_back(0.);
    nMM4.push_back(0.);
    nMM45.push_back(0.);
    nMM5.push_back(0.);
    nTwd.push_back(0.);
    nTwdI.push_back(0.);
    nTrns.push_back(0.);
    nTrnsI.push_back(0.);
    nAway.push_back(0.);
    nAwayI.push_back(0.);
  }

  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    Pythia* pythiaPtr;
    double eCM;

    // Loop over the two different generators
    for (int iGen = 0; iGen <= 3; ++iGen) {

      // Set event-generator pointer
      if (iGen == 0) {
	pythiaPtr = &pythiaA;
	eCM       = eCMa;
      } else if (iGen == 1) {
	pythiaPtr = &pythiaB;
	eCM       = eCMb;
      } else if (iGen == 2) {
	pythiaPtr = &pythiaC;
	eCM       = eCMc;
      } else {
	pythiaPtr = &pythiaD;
	eCM       = eCMd;
      }

      // Shorthand for event record
      Event& event = pythiaPtr->event;
      
      // Generate event
      if (!pythiaPtr->next()) continue;
          
      // Flavor of Z decay
      int flavZqq = abs(event[6].id());
      bool isB = (flavZqq == 5) ? true : false;

      // Add final-state particles to FastJet input vector    
      fjInputs.resize(0);
      for (int i=1;i<event.size();i++) {
	
	// Final state only
	if (!event[i].isFinal())        continue;

	// Count final-state charged multiplicity (as cross check)
	if (event[i].isCharged()) {
	  if (iGen == 0) sumNa++;
	  else if (iGen == 1) sumNb++;
	  else if (iGen == 2) sumNc++;
	  else if (iGen == 3) sumNd++;
	}

	// Now decide what to pass to FastJet
	// NOTE: neutrinos are *included* in this list, since ALEPH
	// appears to have corrected their data to include them.
	// Store as input to Fastjet
	fjInputs.push_back( fastjet::PseudoJet (event[i].px(),event[i].py(),
						event[i].pz(),event[i].e() ) ); 
      }
    
      // FastJet
      if (fjInputs.size() == 0) {
	cout << "Error: event with no final state particles" << endl;
	continue;
      }

      // Run Fastjet algorithm    
      fastjet::ClusterSequence clustSeq(fjInputs, *jetDef);

      // For the first event, print the FastJet details
      if (firstEvent) {
	cout << "Ran " << jetDef->description() << endl;
	cout << "Strategy adopted by FastJet was "
	     << clustSeq.strategy_string() << endl << endl;      
	firstEvent = false;
      }

      // Extract merging scales
      double d56   = clustSeq.exclusive_dmerge(5);
      double y56   = clustSeq.exclusive_ymerge(5);
      double d45   = clustSeq.exclusive_dmerge(4);
      double y45   = clustSeq.exclusive_ymerge(4);
      double lny45 = log(y45);
      double d34   = clustSeq.exclusive_dmerge(3);
      double y34   = clustSeq.exclusive_ymerge(3);
      double lny34 = log(y34);
      // 23
      double y23   = clustSeq.exclusive_ymerge(2);
      double lny23 = log(y23);
      // Check the subjets inside each of the two final jets
      vector<fastjet::PseudoJet> dijet = clustSeq.exclusive_jets(2);
      double mMin = min(dijet[0].m(),dijet[1].m());
      double mMax = max(dijet[0].m(),dijet[1].m());
      
      // Exclusive 4-jet quantities
      bool is4jet = false;
      double the12 = 0.0;
      double the13 = 0.0;
      double the14 = 0.0;
      double the23 = 0.0;
      double the24 = 0.0;
      double e4    = 0.0;

      // Allow light-flavor events to be a bit softer than B
      double yCut = 0.003;
      if ( isB ) yCut = 0.01;
      if (y34 > yCut) {      
	is4jet = true;
	vector<fastjet::PseudoJet> quadjet = clustSeq.exclusive_jets(4);	
	// Sort by energy
	vector<fastjet::PseudoJet> quadjetSorted = sorted_by_E(quadjet);
	// Invariant mass of two softest jets
	the12 = jetAngle(quadjetSorted[0],quadjetSorted[1]);
	the13 = jetAngle(quadjetSorted[0],quadjetSorted[2]);
	the14 = jetAngle(quadjetSorted[0],quadjetSorted[3]);
	the23 = jetAngle(quadjetSorted[1],quadjetSorted[2]);
	the24 = jetAngle(quadjetSorted[1],quadjetSorted[3]);
	//the34 = jetAngle(quadjetSorted[2],quadjetSorted[3]);
	//e3    = quadjetSorted[2].e();
	e4    = quadjetSorted[3].e();
      }
      
      // Check if 1->N last clusterings happen on same side
      bool sameSide13 = true;
      bool sameSide14 = true;
      bool sameSide15 = true;
      double d0_12   = clustSeq.exclusive_subdmerge(dijet[0], 1);
      double d1_12   = clustSeq.exclusive_subdmerge(dijet[1], 1);
      // Check if first subclustering in any of the jets equals second global scale
      if (abs(1-d0_12/d34) < 1e-5 || abs(1-d1_12/d34) < 1e-5 ) {
	sameSide13=false;
	sameSide14=false;
	sameSide15=false;
      }
      // Check if first subclustering in any of the jets equals third global scale
      else if (abs(1-d0_12/d45) < 1e-5 || abs(1-d1_12/d45) < 1e-5 ) {
	sameSide14=false;
	sameSide15=false;
      }
      // Check if first subclustering in any of the jets equals fourth global scale
      else if (abs(1-d0_12/d56) < 1e-5 || abs(1-d1_12/d56) < 1e-5 ) {
	sameSide15=false;
      }

      // Fill histograms (include uncertainty variations for B if on)
      // Event weights 
      double evwt = 1.0;

      // Fill Jet resolution histograms with results from FastJet
      if (y34 > 0.003 && y34 > 0.33*y23) {
        double DM=pow2(mMin/mMax);
        VDM[iGen]->Fill(DM,wDM*evwt);
        if (DM < 0.2) nLM2[iGen]++;
        else nMM2[iGen]++;
        if (DM < 0.25) nLM25[iGen]++;
        else nMM25[iGen]++;
        if (DM < 0.3) nLM3[iGen]++;
        else nMM3[iGen]++;
        if (DM < 0.35) nLM35[iGen]++;
        else nMM35[iGen]++;
        if (DM < 0.4) nLM4[iGen]++;
        else nMM4[iGen]++;
        if (DM < 0.45) nLM45[iGen]++;
        else nMM45[iGen]++;
        if (DM < 0.5) nLM5[iGen]++;
        else nMM5[iGen]++;
        double rM2 = (nMM2[iGen] != 0) ? nLM2[iGen]/nMM2[iGen] : 0.;
        double rM25 = (nMM25[iGen] != 0) ? nLM25[iGen]/nMM25[iGen] : 0.;
        double rM3 = (nMM3[iGen] != 0) ? nLM3[iGen]/nMM3[iGen] : 0.;
        double rM35 = (nMM35[iGen] != 0) ? nLM35[iGen]/nMM35[iGen] : 0.;
        double rM4 = (nMM4[iGen] != 0) ? nLM4[iGen]/nMM4[iGen] : 0.;
        double rM45 = (nMM45[iGen] != 0) ? nLM45[iGen]/nMM45[iGen] : 0.;
        double rM5 = (nMM5[iGen] != 0) ? nLM5[iGen]/nMM5[iGen] : 0.;
        double err2M2 = (nMM2[iGen] != 0) ? 
          nLM2[iGen]/pow2(nMM2[iGen]) * (1.+nLM2[iGen]/nMM2[iGen]) : 0.;
        double err2M25 = (nMM25[iGen] != 0) ? 
          nLM25[iGen]/pow2(nMM25[iGen]) * (1.+nLM25[iGen]/nMM25[iGen]) : 0.;
        double err2M3 = (nMM3[iGen] != 0) ? 
          nLM3[iGen]/pow2(nMM3[iGen]) * (1.+nLM3[iGen]/nMM3[iGen]) : 0.;
        double err2M35 = (nMM35[iGen] != 0) ? 
          nLM35[iGen]/pow2(nMM35[iGen]) * (1.+nLM35[iGen]/nMM35[iGen]) : 0.;
        double err2M4 = (nMM4[iGen] != 0) ? 
          nLM4[iGen]/pow2(nMM4[iGen]) * (1.+nLM4[iGen]/nMM4[iGen]) : 0.;
        double err2M45 = (nMM45[iGen] != 0) ? 
          nLM45[iGen]/pow2(nMM45[iGen]) * (1.+nLM45[iGen]/nMM45[iGen]) : 0.;
        double err2M5 = (nMM5[iGen] != 0) ? 
          nLM5[iGen]/pow2(nMM5[iGen]) * (1.+nLM5[iGen]/nMM5[iGen]) : 0.;
        VDMratio[iGen]->SetBinContent(1,rM2);
        VDMratio[iGen]->SetBinContent(2,rM25);
        VDMratio[iGen]->SetBinContent(3,rM3);
        VDMratio[iGen]->SetBinContent(4,rM35);
        VDMratio[iGen]->SetBinContent(5,rM4);
        VDMratio[iGen]->SetBinContent(6,rM45);
        VDMratio[iGen]->SetBinContent(7,rM5);
        VDMratio[iGen]->SetBinError(1,sqrt(err2M2));
        VDMratio[iGen]->SetBinError(2,sqrt(err2M25));
        VDMratio[iGen]->SetBinError(3,sqrt(err2M3));
        VDMratio[iGen]->SetBinError(4,sqrt(err2M35));
        VDMratio[iGen]->SetBinError(5,sqrt(err2M4));
        VDMratio[iGen]->SetBinError(6,sqrt(err2M45));
        VDMratio[iGen]->SetBinError(7,sqrt(err2M5));
      }
      if (sameSide13 && y34 > 0.003) {
	VLnY24SS[iGen]->Fill(lny34-lny23,wLnY24SS*evwt);
      }
      if (sameSide14 && y45 > 0.003) {
	VLnY25SS[iGen]->Fill(lny45-lny23,wLnY25SS*evwt);
      }
      if (sameSide15 && y56 > 0.003) VSS[iGen]->Fill(4.,evwt);
      else if (sameSide14 && y45 > 0.003) VSS[iGen]->Fill(3.,evwt);
      else if (sameSide13 && y34 > 0.003) VSS[iGen]->Fill(2.,evwt);
      else VSS[iGen]->Fill(1.,evwt);
      
      // Exclusive 4-jet quantities
      if (is4jet) {
	Vx4[iGen]->Fill(2.*e4/eCM,wx4*evwt);
	// Something approaching Simon's color-coherence variable
	// angle of 4th jet with respect to a collinear 3-parton conf
	// Can still be 3-parton soft 
	if (the12 > 2.1 && the13 > 2.1 && the23 < 0.524) {
	  Vthe14[iGen]->Fill(the14/M_PI*180.,wthe14*evwt);
	  Vcos14[iGen]->Fill(cos(the14),wcos14*evwt);
          ValphaStar[iGen]->Fill((abs(the24)-abs(the23))/M_PI,walphaStar*evwt);
          double cos14 = cos(the14);
          if (cos14 > 0.5) nTwd[iGen]++;
          else if (cos14 > -0.5) nTrns[iGen]++;
          else nAway[iGen]++;
          if (cos14 > 0.7) nTwdI[iGen]++;
          else if (abs(cos14) < 0.3) nTrnsI[iGen]++;
          else if (cos14 < -0.7) nAwayI[iGen]++;
          if (nTwd[iGen] > 0) {
            VNF[iGen]->SetBinContent(1,nAway[iGen]/nTwd[iGen]);
            VNF[iGen]->SetBinContent(2,nAway[iGen]/nTwdI[iGen]);
            VNF[iGen]->SetBinContent(3,nTrns[iGen]/(nTwd[iGen]+nAway[iGen]));
            VNF[iGen]->SetBinContent(4,nTrnsI[iGen]/(nTwd[iGen]+nAway[iGen]));
            VNF[iGen]->SetBinContent(5,nTrnsI[iGen]/(nTwdI[iGen]+nAwayI[iGen]));
            double err2AT = 
              nAway[iGen]/pow2(nTwd[iGen])*(1.+nAway[iGen]/nTwd[iGen]);
            double err2ATI = 
              nAwayI[iGen]/pow2(nTwdI[iGen])*(1.+nAwayI[iGen]/nTwdI[iGen]);
            double err2CF = nTrns[iGen]/pow2(nAway[iGen]+nTwd[iGen])
              *(1.+nAway[iGen]/(nAway[iGen]+nTwd[iGen]));
            double err2CIFI = nTrnsI[iGen]/pow2(nAwayI[iGen]+nTwdI[iGen])
              *(1.+nTrnsI[iGen]/(nAwayI[iGen]+nTwdI[iGen]));
            double err2CIF = nTrnsI[iGen]/pow2(nAway[iGen]+nTwd[iGen])
              *(1.+nTrnsI[iGen]/(nAway[iGen]+nTwd[iGen]));
            VNF[iGen]->SetBinError(1,sqrt(err2AT));
            VNF[iGen]->SetBinError(2,sqrt(err2ATI));
            VNF[iGen]->SetBinError(3,sqrt(err2CF));
            VNF[iGen]->SetBinError(4,sqrt(err2CIF));
            VNF[iGen]->SetBinError(5,sqrt(err2CIFI));
          }
	}
      }
    }

    // Periodically update displays    
    vroot.updateDisplay();
    
  }

  // 
  cout<<"------------------------------------------------------------------------\n";
  cout<<"END-OF-RUN GENERATOR A"<<endl;
  cout<<"------------------------------------------------------------------------\n";

  pythiaA.stat();
  // Print out Vincia internal histograms
  if (vinciaOnA) {
    if (verbose >= 2) vinciaA.printHistos();
    vinciaA.printInfo();
  }

  cout<<"------------------------------------------------------------------------\n";
  cout<<"END-OF-RUN GENERATOR B"<<endl;
  cout<<"------------------------------------------------------------------------\n";

  pythiaB.stat();
  // Print out Vincia internal histograms
  if (vinciaOnB) {
    if (verbose >= 2) vinciaB.printHistos();
    vinciaB.printInfo();
  }

  cout<<"------------------------------------------------------------------------\n";
  cout<<"END-OF-RUN GENERATOR C"<<endl;
  cout<<"------------------------------------------------------------------------\n";

  pythiaC.stat();
  // Print out Vincia internal histograms
  if (vinciaOnC) {
    if (verbose >= 2) vinciaC.printHistos();
    vinciaC.printInfo();
  }

  cout<<"------------------------------------------------------------------------\n";
  cout<<"END-OF-RUN GENERATOR D"<<endl;
  cout<<"------------------------------------------------------------------------\n";

  pythiaD.stat();
  // Print out Vincia internal histograms
  if (vinciaOnD) {
    if (verbose >= 2) vinciaD.printHistos();
    vinciaD.printInfo();
  }

  cout << endl << "Enter q to terminate..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
