// vincia31.cc is a part of the VINCIA plugin to PYTHIA 8
// This test program checks the MG5 interface
// Copyright (c) 2017 Peter Skands

// PYTHIA and VINCIA include files and namespaces
#include "Vincia/Vincia.h"

using namespace Pythia8;
using namespace Vincia;

// Main Program
int main() {
  
  //************************************************************************
  // Define Pythia 8 generator with plug-in
  
  Pythia pythia;
  VinciaPlugin vincia(&pythia);
  
  //************************************************************************

  // Shorthands

  MG5interface* mg5ptr = &vincia.mg5interface;  

  //************************************************************************

  // Test adding a user-defined matrix element to the interface
  // (Matrix element must be derived externally and converted with
  // interfaces/MG5/convert.sh script, with .cc and .f files placed in
  // interfaces/MG5/ and .h file placed in include/VinciaMG5/ )
  
  MG5matrix* myAmp=new MG5vincia31_uxu_gg();
  vincia.mg5interface.addMG5matrixPtr(myAmp);

  //************************************************************************

  // Define process type to make init() happy 
  pythia.readString("HardQCD:gg2gg = on");
  pythia.readString("PhaseSpace:pTHatMin = 100.");
  // Switch off MPI and hadronization
  pythia.readString("PartonLevel:MPI = off");
  pythia.readString("HadronLevel:all = off");

  //************************************************************************
  
  // Initialize
  vincia.init();

  //************************************************************************

  // --------

  // 1) Construct a brute-force process gg_gg and check |M|^2 directly

  cout<< "-----"<<endl<<"Test 1 g g > g g (via mg5ptr->getMG5matrix())"<<endl;
  // Create a testLabel (no need to order by PDG code since all = 21)
  vector<int> testLabel;
  testLabel.push_back(21);
  testLabel.push_back(21);
  testLabel.push_back(21);
  testLabel.push_back(21);
  MG5matrix* ampPtr=mg5ptr->getMG5matrixPtr(testLabel);
  if (ampPtr != 0) {
    cout<<"  OK: got MG5matrix* for "<<ampPtr->name()
	<<" with nLegs = "<<ampPtr->nLegs()<<", processLabel = "
	<<mg5ptr->processLabel2string(ampPtr->processList()[0])<<endl;
  } else {
    cout<<"  ERROR! no valid MG5matrix* returned"<<endl;
  }

  // Try with an actual particle configuration
  // (Define as already being in Vincia colour order)
  vector<int> testHel;
  testHel.push_back(1);
  testHel.push_back(1);
  testHel.push_back(1);
  testHel.push_back(1);
  int iA = 0;
  int iB = 1;
  vector<Particle> partons;
  partons.push_back(Particle(testLabel[0],1,0,0,0,0,0,0,
			     0.,0.,1.,1.,0.,0.,testHel[0]));
  partons.push_back(Particle(testLabel[1],1,0,0,0,0,0,0,
			     0.,0.,-1.,1.,0.,0.,testHel[1]));
  partons.push_back(Particle(testLabel[2],1,0,0,0,0,0,0,
			     0.,1.,0.,1.,0.,0.,testHel[2]));
  partons.push_back(Particle(testLabel[3],1,0,0,0,0,0,0,
			     0.,-1.,0.,1.,0.,0.,testHel[3]));
  ampPtr->setPerm(partons,iA,iB);
  ampPtr->setMom(partons);
  int iHel = ampPtr->setHel(testHel);
  if ( iHel < 0 ) {
    cout<<"  ERROR! could not map helicity configuration"<<endl;
  } else {
    cout<<"  OK: got IHEL = "<<iHel<<endl;
  }

  // Set colour configuration
  vector<int> colChain;
  vector< vector<int> > ccList;
  colChain.push_back(0);
  colChain.push_back(1);
  colChain.push_back(3);
  colChain.push_back(2);
  ccList.push_back(colChain);
  int iCol = ampPtr->setIcol(ccList);
  if ( iCol < 0 ) {
    cout<<"  ERROR! could not map colour configuration"<<endl;
  } else {
    cout<<"  OK: got ICOL = "<<iCol<<endl;
  }

  // Get amplitude squared
  double me0 = ampPtr->getME2(0);
  double me1 = ampPtr->getME2(1);
  double me2 = ampPtr->getME2(2);
  cout<<"  MG5 |M|^2 :  LC = "<<me0<<"  VC = "<<me1<<"  FC/IDEN = "<<me2<<endl;
  
  // What if we hadn't put the incoming particles as 0 and 1?
  // (but still use Vincia colour order)
  testHel.resize(0);
  testHel.push_back(1);
  testHel.push_back(1);
  testHel.push_back(1);
  testHel.push_back(1);
  iA = 1;
  iB = 3;
  partons.resize(0);
  partons.push_back(Particle(testLabel[0],1,0,0,0,0,0,0,
			     0.,1.,0.,1.,0.,0.,testHel[0]));
  partons.push_back(Particle(testLabel[1],1,0,0,0,0,0,0,
			     0.,0.,1.,1.,0.,0.,testHel[1]));
  partons.push_back(Particle(testLabel[2],1,0,0,0,0,0,0,
			     0.,-1.,0.,1.,0.,0.,testHel[2]));
  partons.push_back(Particle(testLabel[3],1,0,0,0,0,0,0,
			     0.,0.,-1.,1.,0.,0.,testHel[3]));
  ampPtr->setPerm(partons,iA,iB);
  ampPtr->setMom(partons);
  iHel = ampPtr->setHel(testHel);
  if ( iHel < 0 ) {
    cout<<"  ERROR! could not map helicity configuration"<<endl;
  } else {
    cout<<"  OK: got IHEL = "<<iHel<<endl;
  }

  // Set colour configuration
  ccList.resize(0);
  colChain.resize(0);
  colChain.push_back(0);
  colChain.push_back(1);
  colChain.push_back(2);
  colChain.push_back(3);
  ccList.push_back(colChain);
  iCol = ampPtr->setIcol(ccList);
  if ( iCol < 0 ) {
    cout<<"  ERROR! could not map colour configuration"<<endl;
  } else {
    cout<<"  OK: got ICOL = "<<iCol<<endl;
  }

  // Get amplitude squared
  me0 = ampPtr->getME2(0);
  me1 = ampPtr->getME2(1);
  me2 = ampPtr->getME2(2);
  cout<<"  MG5 |M|^2 :  LC = "<<me0<<"  VC = "<<me1<<"  FC/IDEN = "<<me2<<endl;

  // --------  

  // 2) now try gg_dxd (i.e., nontrivial mapping from Vincia to MG
  //   colour structures)
  // Vincia colour order: d3 g0 g1 dbar2

  cout<< "-----"<<endl<<"Test 2 Nontrivial colour mapping: g g > dx d"<<endl;
  vector<int> testID; testID.resize(4);
  testID[0] = 1;
  testID[1] = 21;
  testID[2] = 21;
  testID[3] = -1;
  iA = 1;
  iB = 2;
  // These labels are given in Vincia colour order. Need to reorder for ME.
  testLabel = mg5ptr->makeProcessLabel(testID, iA, iB);
  cout<<"  made process label "<<mg5ptr->processLabel2string(testLabel)<<endl;
  ampPtr=mg5ptr->getMG5matrixPtr(testLabel);
  if (ampPtr != 0) {
    cout<<"  OK: got MG5matrix* for "<<ampPtr->name()
	<<" with nLegs = "<<ampPtr->nLegs()<<", processLabel = "
	<<mg5ptr->processLabel2string(ampPtr->processList()[0])<<endl;
  } else {
    cout<<"  ERROR! no valid MG5matrix* returned"<<endl;
  }

  // Set momenta (still using Vincia ordering)
  testHel[0] = 1;
  testHel[1] = 1;
  testHel[2] = -1;
  testHel[3] = -1;
  partons.resize(0);
  partons.push_back(Particle(testID[0],1,0,0,0,0,0,0,
			     0.,1.,0.,1.,0.,0.,testHel[0]));
  partons.push_back(Particle(testID[1],1,0,0,0,0,0,0,
			     0.,0.,1.,1.,0.,0.,testHel[1]));
  partons.push_back(Particle(testID[2],1,0,0,0,0,0,0,
			     0.,0.,-1.,1.,0.,0.,testHel[2]));
  partons.push_back(Particle(testID[3],1,0,0,0,0,0,0,
			     0.,-1.,0.,1.,0.,0.,testHel[3]));
  ampPtr->setPerm(partons,iA,iB);
  ampPtr->setMom(partons);
  iHel = ampPtr->setHel(testHel);
  if ( iHel < 0 ) {
    cout<<"  ERROR! could not map helicity configuration"<<endl;
  } else {
    cout<<"  OK: got IHEL = "<<iHel<<endl;
  }
  
  // Set colour configuration
  colChain.resize(0);
  ccList.resize(0);
  colChain.push_back(0);
  colChain.push_back(1);
  colChain.push_back(2);
  colChain.push_back(3);
  ccList.push_back(colChain);
  iCol = ampPtr->setIcol(ccList);
  if ( iCol < 0 ) {
    cout<<"  ERROR! could not map colour configuration"<<endl;
  } else {
    cout<<"  OK: got ICOL = "<<iCol<<endl;
  }

  // Get amplitude squared
  me0 = ampPtr->getME2(0);
  me1 = ampPtr->getME2(1);
  me2 = ampPtr->getME2(2);
  cout<<"  MG5 |M|^2 :  LC = "<<me0<<"  VC = "<<me1<<"  FC/IDEN = "<<me2<<endl;
  
  // --------
  
  // 3) now try dxg_dxg (i.e., nontrivial mapping from Vincia to MG
  //   colour structures)
  // Vincia colour order: dx0 g1 g3 dbar2

  cout<< "-----"<<endl<<"Test 3 Nontrivial colour mapping: dx g > dx g"<<endl;
  testID.resize(4);
  testID[0] = -1;
  testID[1] = 21;
  testID[2] = 21;
  testID[3] = -1;
  iA = 0;
  iB = 1;
  // These labels are given in Vincia colour order. Need to reorder for ME.
  testLabel = mg5ptr->makeProcessLabel(testID, iA, iB);
  cout<<"  made process label "<<mg5ptr->processLabel2string(testLabel)<<endl;
  ampPtr=mg5ptr->getMG5matrixPtr(testLabel);
  if (ampPtr != 0) {
    cout<<"  OK: got MG5matrix* for "<<ampPtr->name()
	<<" with nLegs = "<<ampPtr->nLegs()<<", processLabel = "
	<<mg5ptr->processLabel2string(ampPtr->processList()[0])<<endl;
  } else {
    cout<<"  ERROR! no valid MG5matrix* returned"<<endl;
  }

  // Set momenta (still using Vincia ordering)
  testHel[0] = -1;
  testHel[1] = 1;
  testHel[2] = 1;
  testHel[3] = -1;
  partons.resize(0);
  partons.push_back(Particle(testID[0],1,0,0,0,0,0,0,
			     0.,0.,1.,1.,0.,0.,testHel[0]));
  partons.push_back(Particle(testID[1],1,0,0,0,0,0,0,
			     0.,0.,-1.,1.,0.,0.,testHel[1]));
  partons.push_back(Particle(testID[2],1,0,0,0,0,0,0,
			     0.,1.,0.,1.,0.,0.,testHel[2]));
  partons.push_back(Particle(testID[3],1,0,0,0,0,0,0,
			     0.,-1.,0.,1.,0.,0.,testHel[3]));
  ampPtr->setPerm(partons,iA,iB);
  ampPtr->setMom(partons);
  iHel = ampPtr->setHel(testHel);
  if ( iHel < 0 ) {
    cout<<"  ERROR! could not map helicity configuration"<<endl;
  } else {
    cout<<"  OK: got IHEL = "<<iHel<<endl;
  }
  
  // Set colour configuration
  colChain.resize(0);
  ccList.resize(0);
  colChain.push_back(0);
  colChain.push_back(1);
  colChain.push_back(2);
  colChain.push_back(3);
  ccList.push_back(colChain);
  iCol = ampPtr->setIcol(ccList);
  if ( iCol < 0 ) {
    cout<<"  ERROR! could not map colour configuration"<<endl;
  } else {
    cout<<"  OK: got ICOL = "<<iCol<<endl;
  }

  // Get amplitude squared
  me0 = ampPtr->getME2(0);
  me1 = ampPtr->getME2(1);
  me2 = ampPtr->getME2(2);
  cout<<"  MG5 |M|^2 :  LC = "<<me0<<"  VC = "<<me1<<"  FC/IDEN = "<<me2<<endl;
  
  // 4) now try dxg_uxdxu (two quark pairs with nontrivial mapping
  // between MG and Vincia colour structures)
  // Consider the following Vincia colour order: u4 g1 dx3 dx0 ux2
  cout<< "-----"<<endl<<"Test 4a two quark pairs (->two contractions): dx g > ux dx u"<<endl;
  testID.resize(5);
  testID[0] = 2;
  testID[1] = 21;
  testID[2] = -1;
  testID[3] = -1;
  testID[4] = -2;
  iA = 3;
  iB = 1;
  // These labels are given in Vincia colour order. Need to reorder for ME.
  testLabel = mg5ptr->makeProcessLabel(testID, iA, iB);
  cout<<"  made process label "<<mg5ptr->processLabel2string(testLabel)<<endl;
  ampPtr=mg5ptr->getMG5matrixPtr(testLabel);
  if (ampPtr != 0) {
    cout<<"  OK: got MG5matrix* for "<<ampPtr->name()
	<<" with nLegs = "<<ampPtr->nLegs()<<", processLabel = "
	<<mg5ptr->processLabel2string(ampPtr->processList()[0])<<endl;
  } else {
    cout<<"  ERROR! no valid MG5matrix* returned"<<endl;
  }

  // Set momenta and helicities (still using Vincia ordering)
  testHel.resize(5);
  testHel[0] = -1;
  testHel[1] = 1;
  testHel[2] = 1;
  testHel[3] = 1;
  testHel[4] = 1;
  partons.resize(0);
  partons.push_back(Particle(testID[0],1,0,0,0,0,0,0,
			     -6.,0.,0.,6.,0.,0.,testHel[0]));
  partons.push_back(Particle(testID[1],1,0,0,0,0,0,0,
			     0.,0.,-8.0,8.,0.,0.,testHel[1]));
  partons.push_back(Particle(testID[2],1,0,0,0,0,0,0,
			     3.,4.,0.,5.,0.,0.,testHel[2]));
  partons.push_back(Particle(testID[3],1,0,0,0,0,0,0,
			     0.,0.,8.,8.,0.,0.,testHel[3]));
  partons.push_back(Particle(testID[4],1,0,0,0,0,0,0,
			     3.,-4.,0.,5.,0.,0.,testHel[4]));
  ampPtr->setPerm(partons,iA,iB);
  ampPtr->setMom(partons);
  iHel = ampPtr->setHel(testHel);
  if ( iHel < 0 ) {
    cout<<"  ERROR! could not map helicity configuration"<<endl;
  } else {
    cout<<"  OK: got IHEL = "<<iHel<<endl;
  }

  // Set colour configuration
  ccList.resize(0);
  colChain.resize(0);
  colChain.push_back(0);
  colChain.push_back(1);
  colChain.push_back(2);
  ccList.push_back(colChain);
  colChain.resize(0);
  colChain.push_back(3);
  colChain.push_back(4);
  ccList.push_back(colChain);
  iCol = ampPtr->setIcol(ccList);
  if ( iCol < 0 ) {
    cout<<"  ERROR! could not map colour configuration"<<endl;
  } else {
    cout<<"  OK: got ICOL = "<<iCol<<endl;
  } 
  
  // --------
  
  // Check that relative order of multiple traces do no matter.
  // Vincia colour traces in opposite order -> same MG trace

  cout<< "-----"<<endl<<"Test 4b same two contractions in opposite order (-> same ICOL, same ME)"<<endl;
  ccList.resize(0);
  colChain.resize(0);
  colChain.push_back(3);
  colChain.push_back(4);
  ccList.push_back(colChain);
  colChain.resize(0);
  colChain.push_back(0);
  colChain.push_back(1);
  colChain.push_back(2);
  ccList.push_back(colChain);
  iCol = ampPtr->setIcol(ccList);
  if ( iCol < 0 ) {
    cout<<"  ERROR! could not map colour configuration"<<endl;
  } else {
    cout<<"  OK: got ICOL = "<<iCol<<endl;
  } 

  // Get amplitude squared
  me0 = ampPtr->getME2(0);
  me1 = ampPtr->getME2(1);
  me2 = ampPtr->getME2(2);
  cout<<"  MG5 |M|^2 :  LC = "<<me0<<"  VC = "<<me1<<"  FC/IDEN = "<<me2<<endl;

  // --------
  
  // 5) see if we can access the user-defined matrix element for uxu_gg

  cout<< "-----"<<endl<<"Test 5 accessing a user-added ME for uxu_gg"<<endl;
  testID.resize(4);
  testID[0] = -2;
  testID[1] = 21;
  testID[2] = 21;
  testID[3] = 2;
  iA = 3;
  iB = 0;
  // These labels are given in Vincia colour order. Need to reorder for ME.
  testLabel = mg5ptr->makeProcessLabel(testID, iA, iB);
  cout<<"  made process label "<<mg5ptr->processLabel2string(testLabel)<<endl;
  ampPtr=mg5ptr->getMG5matrixPtr(testLabel);
  if (ampPtr != 0) {
    cout<<"  OK: got MG5matrix* for "<<ampPtr->name()
	<<" with nLegs = "<<ampPtr->nLegs()<<", processLabel = "
	<<mg5ptr->processLabel2string(ampPtr->processList()[0])<<endl;
  } else {
    cout<<"  ERROR! no valid MG5matrix* returned"<<endl;
  }

  // Set momenta (still using Vincia ordering)
  testHel.resize(4);
  testHel[0] = 1;
  testHel[1] = -1;
  testHel[2] = 1;
  testHel[3] = -1;
  partons.resize(0);
  partons.push_back(Particle(testID[0],1,0,0,0,0,0,0,
			     0.,0.,1.,1.,0.,0.,testHel[0]));
  partons.push_back(Particle(testID[1],1,0,0,0,0,0,0,
			     0.,1.,0.,1.,0.,0.,testHel[1]));
  partons.push_back(Particle(testID[2],1,0,0,0,0,0,0,
			     0.,-1.,0.,1.,0.,0.,testHel[2]));
  partons.push_back(Particle(testID[3],1,0,0,0,0,0,0,
			     0.,0.,-1.,1.,0.,0.,testHel[3]));
  ampPtr->setPerm(partons,iA,iB);
  ampPtr->setMom(partons);
  iHel = ampPtr->setHel(testHel);
  if ( iHel < 0 ) {
    cout<<"  ERROR! could not map helicity configuration"<<endl;
  } else {
    cout<<"  OK: got IHEL = "<<iHel<<endl;
  }
  
  // Set colour configuration
  colChain.resize(0);
  ccList.resize(0);
  colChain.push_back(0);
  colChain.push_back(1);
  colChain.push_back(2);
  colChain.push_back(3);
  ccList.push_back(colChain);
  iCol = ampPtr->setIcol(ccList);
  if ( iCol < 0 ) {
    cout<<"  ERROR! could not map colour configuration"<<endl;
  } else {
    cout<<"  OK: got ICOL = "<<iCol<<endl;
  }

  // Get amplitude squared
  me0 = ampPtr->getME2(0);
  me1 = ampPtr->getME2(1);
  me2 = ampPtr->getME2(2);
  cout<<"  MG5 |M|^2 :  LC = "<<me0<<"  VC = "<<me1<<"  FC/IDEN = "<<me2<<endl;
  
  // --------
  
  // Time to see if we can interface any of these matrix elements via the
  // MG5 interface rather than directly. Use gg_ggg as first example.

  cout<< "-----"<<endl<<"Test 6 using MG5interface::ME2() function"<<endl;
  // Create a list of particles such as might be produced by Vincia,
  // still assuming they've already been put in Vincia colour order  
  partons.resize(5);
  partons[0]=Particle(21, 23,0,0,0,0,0,0, -6., 0., 0., 6.,0.,0.,9);
  partons[1]=Particle(21,-21,0,0,0,0,0,0,  0., 0.,-8., 8.,0.,0.,9);
  partons[2]=Particle(21, 23,0,0,0,0,0,0,  3., 4., 0., 5.,0.,0.,9);
  partons[3]=Particle(21,-21,0,0,0,0,0,0,  0., 0., 8., 8.,0.,0.,9);
  partons[4]=Particle(21, 23,0,0,0,0,0,0,  3.,-4., 0., 5.,0.,0.,9);
  iA = 1;
  iB = 3;
  double ME2 = mg5ptr->ME2(partons, iA, iB);
  cout<<"  MG5 |M|^2 = "<<ME2<<endl;

  // --------

  // Now see if we can use the selectHelicities function to set
  // helicities for an initially inpolarised state.
  // (Just use same state as in previous example)
 
  cout<< "-----"<<endl<<"Test 7 Helicity Selection"<<endl;
  cout<<"   helicities before selection:";
  for (int i=0; i<(int)partons.size(); ++i)
    cout<<" "<<num2str(partons[i].pol());
  cout<<endl;
  bool ok = mg5ptr->selectHelicities( partons, iA, iB );
  cout<<"   helicities after selection: ";
  for (int i=0; i<(int)partons.size(); ++i)
    cout<<" "<<num2str(partons[i].pol());
  cout<<endl;

 // --------

  // Now see if we can use the colourSort function to bring a system
  // of particles into Vincia colour order
 
  cout<< "-----"<<endl<<"Test 8 Colour Sorting for gluon ring"<<endl;
  partons[0]=Particle(21, 23,0,0,0,0,501,502, -6., 0., 0., 6.,0.,0.,9);
  partons[1]=Particle(21,-21,0,0,0,0,503,502,  0., 0.,-8., 8.,0.,0.,9);
  partons[2]=Particle(21, 23,0,0,0,0,503,504,  3., 4., 0., 5.,0.,0.,9);
  partons[3]=Particle(21,-21,0,0,0,0,505,504,  0., 0., 8., 8.,0.,0.,9);
  partons[4]=Particle(21, 23,0,0,0,0,505,501,  3.,-4., 0., 5.,0.,0.,9);
  vector<Particle*> particlePtrs;
  particlePtrs.resize(partons.size());
  particlePtrs[0]=&partons[0];
  particlePtrs[1]=&partons[1];
  particlePtrs[2]=&partons[2];
  particlePtrs[3]=&partons[3];
  particlePtrs[4]=&partons[4];
  vector<int> iVinciaOrder = vincia.vinCom.colourSort(particlePtrs);
  cout<<" order 2 is";
  for (int i=0; i<(int)iVinciaOrder.size(); ++i)
    cout<<" "<<iVinciaOrder[i]<<"("<<partons[iVinciaOrder[i]].id()<<(partons[iVinciaOrder[i]].isFinal()?"":"x")<<") ";
  cout<<endl;
    
  partons[0]=Particle(21, 23,0,0,0,0,503,504, -6., 0., 0., 6.,0.,0.,9);
  partons[1]=Particle(21,-21,0,0,0,0,503,502,  0., 0.,-8., 8.,0.,0.,9);
  partons[2]=Particle(21, 23,0,0,0,0,501,502,  3., 4., 0., 5.,0.,0.,9);
  partons[3]=Particle(21,-21,0,0,0,0,505,504,  0., 0., 8., 8.,0.,0.,9);
  partons[4]=Particle(21, 23,0,0,0,0,505,501,  3.,-4., 0., 5.,0.,0.,9);
  particlePtrs.resize(partons.size());
  particlePtrs[0]=&partons[0];
  particlePtrs[1]=&partons[1];
  particlePtrs[2]=&partons[2];
  particlePtrs[3]=&partons[3];
  particlePtrs[4]=&partons[4];
  iVinciaOrder = vincia.vinCom.colourSort(particlePtrs);
  cout<<" order 2 is";
  for (int i=0; i<(int)iVinciaOrder.size(); ++i)
    cout<<" "<<iVinciaOrder[i]<<"("<<partons[iVinciaOrder[i]].id()<<(partons[iVinciaOrder[i]].isFinal()?"":"x")<<") ";
  cout<<endl;
      
  cout<< "-----"<<endl<<"Test 9 Colour Sorting for qqbar + gluons"<<endl;
  cout<< "(should start on FS quark (IS antiquark) and end on FS antiquark (or IS quark); x denotes IS (crossing)"<<endl;
  partons[0]=Particle( 1, 23,0,0,0,0,501,  0, -6., 0., 0., 6.,0.,0.,9);
  partons[1]=Particle( 1,-21,0,0,0,0,503,  0,  0., 0.,-8., 8.,0.,0.,9);
  partons[2]=Particle(21, 23,0,0,0,0,503,504,  3., 4., 0., 5.,0.,0.,9);
  partons[3]=Particle(21,-21,0,0,0,0,505,504,  0., 0., 8., 8.,0.,0.,9);
  partons[4]=Particle(21, 23,0,0,0,0,505,501,  3.,-4., 0., 5.,0.,0.,9);  
  particlePtrs.resize(partons.size());
  particlePtrs[0]=&partons[0];
  particlePtrs[1]=&partons[1];
  particlePtrs[2]=&partons[2];
  particlePtrs[3]=&partons[3];
  particlePtrs[4]=&partons[4];
  iVinciaOrder = vincia.vinCom.colourSort(particlePtrs);
  cout<<" order 1 is";
  for (int i=0; i<(int)iVinciaOrder.size(); ++i)
    cout<<" "<<iVinciaOrder[i]<<"("<<partons[iVinciaOrder[i]].id()<<(partons[iVinciaOrder[i]].isFinal()?"":"x")<<") ";
  cout<<endl;
  
  partons[0]=Particle(21, 23,0,0,0,0,501,502, -6., 0., 0., 6.,0.,0.,9);
  partons[1]=Particle(-1,-21,0,0,0,0,  0,502,  0., 0.,-8., 8.,0.,0.,9);
  partons[2]=Particle(-1, 23,0,0,0,0,  0,504,  3., 4., 0., 5.,0.,0.,9);
  partons[3]=Particle(21,-21,0,0,0,0,505,504,  0., 0., 8., 8.,0.,0.,9);
  partons[4]=Particle(21, 23,0,0,0,0,505,501,  3.,-4., 0., 5.,0.,0.,9);  
  particlePtrs.resize(partons.size());
  particlePtrs[0]=&partons[0];
  particlePtrs[1]=&partons[1];
  particlePtrs[2]=&partons[2];
  particlePtrs[3]=&partons[3];
  particlePtrs[4]=&partons[4];
  iVinciaOrder = vincia.vinCom.colourSort(particlePtrs);
  cout<<" order 2 is";
  for (int i=0; i<(int)iVinciaOrder.size(); ++i)
    cout<<" "<<iVinciaOrder[i]<<"("<<partons[iVinciaOrder[i]].id()<<(partons[iVinciaOrder[i]].isFinal()?"":"x")<<") ";
  cout<<endl;

  cout<< "-----"<<endl<<"Test 10 Colour Sorting for two chains"<<endl;
  // Put initial-state partons first in this list, then do colour sort
  partons[0]=Particle(21,-21,0,0,0,0,501,505,  0., 0.,-8., 8.,0.,0.,9);
  partons[1]=Particle(-2,-21,0,0,0,0,  0,501,  0., 0., 8., 8.,0.,0.,9);
  partons[2]=Particle( 1, 23,0,0,0,0,503,  0, -6., 0., 0., 6.,0.,0.,9);
  partons[3]=Particle(-1, 23,0,0,0,0,  0,505,  3., 4., 0., 5.,0.,0.,9);
  partons[4]=Particle(-2, 23,0,0,0,0,  0,503,  3.,-4., 0., 5.,0.,0.,9);  
  iA = 0;
  iB = 1;
  particlePtrs.resize(partons.size());
  particlePtrs[0]=&partons[0];
  particlePtrs[1]=&partons[1];
  particlePtrs[2]=&partons[2];
  particlePtrs[3]=&partons[3];
  particlePtrs[4]=&partons[4];
  iVinciaOrder = vincia.vinCom.colourSort(particlePtrs);
  cout<<" order is";
  for (int i=0; i<(int)iVinciaOrder.size(); ++i)
    cout<<" "<<iVinciaOrder[i]<<"("<<partons[iVinciaOrder[i]].id()<<(partons[iVinciaOrder[i]].isFinal()?"":"x")<<") ";
  cout<<endl;

  // See if we can get a matrix element for this state
  vector<Particle> partonsOrdered;
  for (int i=0; i<iVinciaOrder.size(); ++i)
    partonsOrdered.push_back(partons[iVinciaOrder[i]]);
  int iAordered = iVinciaOrder[iA];
  int iBordered = iVinciaOrder[iB];
  ME2 = mg5ptr->ME2(partonsOrdered, iAordered, iBordered);
  cout<<"  MG5 |M|^2 = "<<ME2<<endl;
  
  // Done.
  return 0;
}
