// MergingHooks.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2017 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the VinciaMergingHooks class.
// Most of the code is copied from the Pythia MergingHooks class.

#ifndef Vincia_MergingHooks_H
#define Vincia_MergingHooks_H

#include "Pythia8/MergingHooks.h"
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"

using namespace Pythia8;

namespace Vincia {

  // Forward declarations.
  class TimeAntShower;
  class SpaceAntShower;
  class VinciaMerging;

  class VinciaMergingHooks : public Pythia8::MergingHooks {

  public:

  // Constructor.
  VinciaMergingHooks() { 
      inputEvent        = Event();
      resonances.resize(0); 
      settingsPtr       = 0;
      infoPtr           = 0;
      particleDataPtr   = 0;
      partonSystemsPtr  = 0;
      timesAntShowerPtr = 0;
      spaceAntShowerPtr = 0;
      doPTLundMergingSave = false;
      useOrderedPathSave  = false;
      useCompletePathSave = false;
      doOrderedMECsSave   = false;
    }

  // Destructor.
  ~VinciaMergingHooks() {}

  // Set the necessary Vincia pointers.
  void setVinciaPtrs(TimeAntShower* tPtrIn, SpaceAntShower* sPtrIn,
    VinciaCommon* cPtrIn) {
    timesAntShowerPtr = tPtrIn;
    spaceAntShowerPtr = sPtrIn;
    vinComPtr         = cPtrIn;
  }

  // Initialize.
  void init();

  // Set starting scales.
  bool setShowerStartingScales(bool isTrial, bool doMergeFirstEmm, 
    double& pTscaleIn, const Event& event,
    double& pTmaxFSRIn, bool& limitPTmaxFSRIn,
    double& pTmaxISRIn, bool& limitPTmaxISRIn,
    double& pTmaxMPIIn, bool& limitPTmaxMPIIn );

  // Calculate the merging scale for the input event.
  double tmsCalc(const Event& event);
  double rhoVincia(const Event& event, int rad, int emt, int rec, int type);
  double tmsNow(const Event& event) { return tmsCalc(event); }

  // Function to calculate the hard process matrix element.
  double hardProcessME(const Event& event);

  // Output functions.

  // Function returning the value of the merging scale.
  double tms() { return tmsValueSave; }

  // Functions to determine the merging procedure definition.
  bool doPTLundMerging() { return doPTLundMergingSave; }
  bool doCKKWLMerging()  { return doPTLundMergingSave; }

  // Return whether cross section is estimated.
  bool doEstimateXSec() { return doEstimateXSection; }

  // Return number of requested additional jets on top of the Born process.
  int nRequested() { return nRequestedSave; }

  // Prefactor for variation of effective scale
  double unorderedScaleVariation() { return unorderedScaleVariationSave; }

  // Return whether MECs should be performed in the histories.
  bool doMECs()             { return doMECsSave; }
  // Return whether only shower-reproducible events should be included.
  bool showerPathsOnly()    { return showerPathsOnlySave; }
  // Return whether only non-shower-reproducible events should be included.
  bool nonShowerPathsOnly() { return nonShowerPathsOnlySave; }
  // Return whether cutoff in non-shower-reproducible events should be used.
  bool checkShowerCutoff()  { return checkShowerCutoffSave; }
  double showerCutoff()     { return showerCutoffSave; }

  // Return coupling objects
  AlphaStrong* AlphaS_FSR() { return &vinComPtr->alphaS; }
  AlphaStrong* AlphaS_ISR() { return &vinComPtr->alphaS; }
  AlphaEM* AlphaEM_FSR()    { return &vinComPtr->alphaEM; }
  AlphaEM* AlphaEM_ISR()    { return &vinComPtr->alphaEM; }

  // Output and input functions for communication with the shower.
  bool useShowerPlugin() { return true; }
  // Functions to set and return whether a chosen path is ordered.
  void useOrderedPath(bool ordered) { useOrderedPathSave = ordered; }
  bool useOrderedPath()             { return useOrderedPathSave; }
  // Functions to set and return whether a chosen path is complete.
  void useCompletePath(bool complete) { useCompletePathSave = complete; }
  bool useCompletePath()              { return useCompletePathSave; }
  // Functions to set and return whether strong MECs can be performed.
  void doOrderedMECs(bool correct) { doOrderedMECsSave = correct; }
  bool doOrderedMECs()             { return doOrderedMECsSave; }
  // Functions so set and return how many good histories exist.
  void nrGoodBranch(int nGood) { nrGoodBranchSave = nGood; }
  int  nrGoodBranch()          { return nrGoodBranchSave; }

  // Pointer to Vincia objects.
  TimeAntShower*  timesAntShowerPtr;
  SpaceAntShower* spaceAntShowerPtr;
  VinciaCommon*   vinComPtr;

  // Flags for merging procedure definition.
  bool   doMECsSave, showerPathsOnlySave, nonShowerPathsOnlySave,
         checkShowerCutoffSave;
  double showerCutoffSave, unorderedScaleVariationSave;

  // Parameters for communication with the shower.
  bool   useOrderedPathSave, useCompletePathSave, doOrderedMECsSave;
  int    nrGoodBranchSave;

  };

}

#endif
