// VinciaRoot.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Vincia ROOT utilities

#ifndef VINCIAROOT_H
#define VINCIAROOT_H

// Include the Pythia Settings header
#include "Pythia8/Settings.h"

// Include the VinciaRoot headers 
// (ROOT and STDLIB headers included in VinciaSettings.h)

#include "VinciaRoot/VRdisplay.h"
#include "VinciaRoot/VRfile.h"
#include "VinciaRoot/VRgraph.h"
#include "VinciaRoot/VRsettings.h"

// Put all the rest in Vincia namespace

namespace Vincia {
  
  //***************************************************************
  // class VinciaRoot: Main VinciaRoot Class
  // Note: is completely independent of VINCIA and can be used 
  // with any executable using ROOT histograms (specifically TH1F)
  //***************************************************************

class VinciaRoot : public TApplication {

  public:

    // Constructor (requires a pointer to a Pythia 8 settings database)
    VinciaRoot(Pythia8::Settings* settingsPtrIn)
      : TApplication("VinciaRootApp",0,0), nUpdateCallsSav(-1.) {

      // Switch on storage of stat uncertainties 
      TH1::SetDefaultSumw2();

      // Check if Root display app has been initialized
      if ( !settings.isInitApp ) initApp(); 

      // Save pointer to Settings database and pass on to VinciaRoot settings
      settingsPtr = settingsPtrIn;
      settings.settingsPtr = settingsPtr;
      
    }

    // Initialise 
    void init() {
      settings.init();
      // Set MC names and versions
      string pyVer = num2str(settingsPtr->parm("Pythia:versionNumber"),5);  
      string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),5);  
      string mcVersions = "Vincia " + vcVer + " + Pythia "+pyVer;      
      setMCnames("all",mcVersions);
    }
    
    // Initialize ROOT app
    void initApp() {
      cout<<"(VinciaRoot): initializing display app"<<endl;
      gROOT->SetStyle("Plain");
      InitializeGraphics();
      settings.isInitApp = true;
      // Set default hatching width thick enough to show up well on slides
      gStyle->SetHatchesLineWidth(2);
      // Speed up drawing, with openGL
      gStyle->SetCanvasPreferGL(kTRUE);
    }

    // Check if a display exists
    bool exists(string displayName) {
      if (displays.find(displayName) == displays.end()) return false;
      else return true;
    }

    // Initialize a display
    void createDisplay(string displayName, int xSizIn=-1, int ySizIn=-1, 
                     int xOffsetIn=-1,int yOffsetIn=-1) {

      // Check if display with that name already exists
      if ( exists(displayName) ) {
        cout<<"(VinciaRoot): warning! a display named "<<displayName
            <<" already exists"<<endl;
        displays.erase(displayName);
      }

      // Use default plot sizes if none specified
      if (xSizIn == -1) xSizIn = settings.plotSizeX;
      if (ySizIn == -1) ySizIn = settings.plotSizeY;

      // Cascade windows from top left corner if no offsets specified
      if (xOffsetIn == -1) xOffsetIn = 64+16*displays.size();
      if (yOffsetIn == -1) yOffsetIn = 32+16*displays.size();

      // Create and initialize display 
      displays[displayName] = new VRdisplay(displayName,xSizIn,ySizIn,
                                            xOffsetIn,yOffsetIn);
      displays[displayName]->init(&settings);
 
    }

    // Create a new Display from TH1D*
    void createDisplay(string  displayName, TH1D* HisPtrIn, 
                       int xSizIn=-1, int ySizIn=-1, 
                       int xOffsetIn=-1,int yOffsetIn=-1) {
      // Initialize display and associate 
      createDisplay(displayName, xSizIn, ySizIn, xOffsetIn, yOffsetIn);
      displays[displayName]->addTH1DPtr(HisPtrIn);
    }
    
    // Create a new Display from vector<TH1D*>
    void createDisplay(string displayName, vector<TH1D*> HisPtrIn, 
                       int xSizIn=-1, int ySizIn=-1, 
                       int xOffsetIn=-1,int yOffsetIn=-1) {
      // Initialize display and associate 
      createDisplay(displayName, xSizIn, ySizIn, xOffsetIn, yOffsetIn);
      displays[displayName]->addTH1DPtr(HisPtrIn);
    }

    // Create a new Display from TH2D*
    void createDisplay(string  displayName, TH2D* HisPtrIn, 
                       int xSizIn=-1, int ySizIn=-1, 
                       int xOffsetIn=-1,int yOffsetIn=-1) {
      // Initialize display and associate 
      createDisplay(displayName, xSizIn, ySizIn, xOffsetIn, yOffsetIn);
      displays[displayName]->setTH2DPtr(HisPtrIn);
    }

    // Set default plot sizes
    void setPlotSize(int plotSizeXin, int plotSizeYin) {
      settings.plotSizeX = plotSizeXin;
      settings.plotSizeY = plotSizeYin;
    }

    // (Re)set global verbose level
    void setVerbose(string displayNameIn, int verboseIn) {
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.verbose = verboseIn;
        // Propagate change to all displays
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setVerbose(verboseIn);
      } 
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") 
        settings.verbose = verboseIn;
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setVerbose(verboseIn);
      else cout<<"(VinciaRoot): setVerbose no display named "
               <<displayNameIn<<endl;
    }
    
    // (Re)set global optstat
    void setOptStat(string displayNameIn = "all", int optStatIn = 1) {
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.optStat = optStatIn;
        // Propagate change to all displays
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setOptStat(optStatIn);
      } 
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") 
        settings.optStat = optStatIn;
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setOptStat(optStatIn);
      else cout<<"(VinciaRoot): setOptStat no display named "
               <<displayNameIn<<endl;
    }

    // Set normalize on/off
    void setNormalize(string displayNameIn="all", 
                      bool normalizeIn = true, double normFacIn = 1.0) { 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.normalize = normalizeIn;
        settings.normFac   = normFacIn;
        // Propagate change to all displays
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setNormalize(normalizeIn,normFacIn);
      } 
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") {
        settings.normalize = normalizeIn;
        settings.normFac   = normFacIn;
      } else if (exists(displayNameIn)) 
        displays[displayNameIn]->setNormalize(normalizeIn,normFacIn);
      else cout<<"(VinciaRoot): setNormalize no display named "
               <<displayNameIn<<endl;
    }

    // Accept British spelling
    void setNormalise(string displayNameIn="all", 
                      bool normalizeIn = true, double normFacIn = 1.0) {
      setNormalize(displayNameIn, normalizeIn, normFacIn);
    }
    
    // Log Y scale 
    void setLogY(string displayNameIn="all", bool logYin=true) {
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.logY = logYin;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setLogY(logYin);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") 
        settings.logY = logYin;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) displays[displayNameIn]->setLogY(logYin);
      else cout<<"(VinciaRoot): setLogY no display named "<<displayNameIn<<endl;
    }

    // Wrapper to set Linear Y scale 
    void setLinY(string displayNameIn="all", bool linYin=true) {
      setLogY(displayNameIn, !linYin);
    }
    
    // Set logarithmic X axis
    void setLogX(string displayNameIn="all", bool logXin=true) {
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.logX = logXin;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setLogX(logXin);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") 
        settings.logX = logXin;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) displays[displayNameIn]->setLogX(logXin);
      else cout<<"(VinciaRoot): setLogX no display named "<<displayNameIn<<endl;
    }    

    // Set linear X axis
    void setLinX(string displayNameIn="all", bool linX=true) { 
      setLogX(displayNameIn, !linX); 
    }

    // Set MC names
    void setMCnames(string displayNameIn, string MCnamesIn) {
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.MCnames = MCnamesIn;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setMCnames(MCnamesIn);
      } 
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
                 displayNameIn == "NEW") 
        settings.MCnames = MCnamesIn;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setMCnames(MCnamesIn);
      else cout<<"(VinciaRoot): setMCnames no display named "
               <<displayNameIn<<endl;
    }

    // Set legend (with optional alignment, default 0 is auto)
    void setLegend(string displayNameIn, vector<string> legendIn, 
                   int alignIn = 0) {
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.legend = legendIn;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setLegend(legendIn);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
                 displayNameIn == "NEW") 
        settings.legend = legendIn;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setLegend(legendIn, alignIn);
      else cout<<"(VinciaRoot): setLegend no display named "
               <<displayNameIn<<endl;
    }

    // Wrapper for passing single legend as string
    void setLegend(string displayNameIn, 
                   string legendIn, int alignIn = 0) { 
      vector<string> legendVec;
      legendVec.push_back(legendIn);
      setLegend(displayNameIn, legendVec,alignIn);
    }

    // Set labels for uncertainty bands
    void setUncertaintyLabels(string displayNameIn, vector<string> labelsIn) {
     // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.uncertaintyLabels = labelsIn;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setUncertaintyLabels(labelsIn);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
                 displayNameIn == "NEW") 
        settings.uncertaintyLabels = labelsIn;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setUncertaintyLabels(labelsIn);
      else cout<<"(VinciaRoot): setUncertaintyLabels no display named "
               <<displayNameIn<<endl;      
    }

    // Set Data Reference (with optional alignment, default 0 is auto)
    void setDataRef(string displayNameIn, string dataRefIn) {
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setDataRef(dataRefIn);
      }
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setDataRef(dataRefIn);
      else cout<<"(VinciaRoot): setDataRef no display named "
               <<displayNameIn<<endl;
    }
    
    // Set whether to show ratio pane by default.
    // Optionally loop over all existing plots as well
    void setShowRatioPad(string displayNameIn = "all", bool showRatioIn=true) { 
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.showRatioPad = showRatioIn;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setShowRatioPad(showRatioIn);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
                 displayNameIn == "NEW") 
        settings.showRatioPad = showRatioIn;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setShowRatioPad(showRatioIn);
      else cout<<"(VinciaRoot): setShowRatioPad no display named "
               <<displayNameIn<<endl;
    }

    // Set whether to show ratio pane by default.
    // Optionally loop over all existing plots as well
    void setShowRelUnc(string displayNameIn = "all", bool showRelUncIn=true) { 
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.showRelUncPads = showRelUncIn;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setShowRelUnc(showRelUncIn);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
                 displayNameIn == "NEW") 
        settings.showRelUncPads = showRelUncIn;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setShowRelUnc(showRelUncIn);
      else cout<<"(VinciaRoot): setShowRelUnc no display named "
               <<displayNameIn<<endl;
    }

    // Set whether to show chi2(5%) values with legends
    void setShowLegendChi2(string displayNameIn = "all", bool showChi2=true) { 
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.showLegendChi2 = showChi2;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setShowLegendChi2(showChi2);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
                 displayNameIn == "NEW") 
        settings.showLegendChi2 = showChi2;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setShowLegendChi2(showChi2);
      else cout<<"(VinciaRoot): setShowRatioPad no display named "
               <<displayNameIn<<endl;
    }
    
    // Set range of ratio pane plot
    void setRatioRange(string displayNameIn, double yMin, double yMax) {
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.ratioMin = yMin;
        settings.ratioMax = yMax;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setRatioRange(yMin,yMax);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") {
        settings.ratioMin = yMin;
        settings.ratioMax = yMax;
      // Else set for a single (named) display
      } else if (exists(displayNameIn)) 
        displays[displayNameIn]->setRatioRange(yMin,yMax);
      else cout<<"(VinciaRoot): setRatioRange no display named "
               <<displayNameIn<<endl;
    }

    // Wrapper to set colour scheme for a specific graph on a specific display
    virtual bool setColorScheme(string displayNameIn, 
                                int iScheme, int iGraph=1) { 
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setColorScheme(iScheme,iGraph);
        return true;
      }
      // Keyword "new" not used in this method
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") {
        cout<<"(VinciaRoot): setColorScheme keyword 'new' not used "
            <<"in this method"<<endl;
        return false;
      // Else set for a single (named) display
      } else if (exists(displayNameIn)) {
        return displays[displayNameIn]->setColorScheme(iScheme,iGraph);
      } else {
        cout<<"(VinciaRoot): setColorScheme no display named "
            <<displayNameIn<<endl;
        return false;
      }
      return false;
    }

    // Functions to give access to member functions of individual displays
    void addTH1DPtr(string displayName, TH1D* hisPtrIn) {
      if ( !exists(displayName) ) createDisplay(displayName, hisPtrIn);
      else displays[displayName]->addTH1DPtr(hisPtrIn);
    }
    void addTH1DPtr(string displayName, vector<TH1D*> hisPtrIn) {
      if ( !exists(displayName) ) createDisplay(displayName, hisPtrIn);
      else displays[displayName]->addTH1DPtr(hisPtrIn);
    }    
    // setDataFile():
    // Set file to be treated as experimental data 
    void setDataFile(string displayName, string fileIn, bool hasSysErrors=true) {
      if (hasSysErrors) setDataFile(displayName,fileIn,10,20,30,40,51,52,61,62);
      else setDataFile(displayName,fileIn,10,20,30,40,51,52);
    }
    void setDataFile(string displayName, string fileIn, 
      int i1, int i2, int i3=-1, int i4=-1, int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, int i10=-1, int i11=-1, int i12=-1, int i13=-1, int i14=-1, int i15=-1, int i16=-1, int i17=-1, int i18=-1, int i19=-1) {
      if ( !exists(displayName) && settings.verbose >= 1) {
        cout<<"(VinciaRoot): ERROR! display "<<displayName
            <<" does not exist. Attempt to add data file failed."<<endl;
        return;
      }
      displays[displayName]->setDataFile(fileIn, i1, i2, i3, i4, i5, i6, i7, i8, i9, i10, i11, i12, i13, i14, i15, i16, i17, i18, i19);
    } 
    // addHisFile() : 
    // Add file with saved histogram
    // (Default is to ignore any additional systematic errors)
    void addHisFile(string displayName, string fileIn,  
                    bool hasSysErrors=false) {
      if (hasSysErrors) addHisFile(displayName,fileIn,10,20,30,40,51,52,61,62);
      else addHisFile(displayName,fileIn,10,20,30,40,51,52);
    }
    void addHisFile(string displayName,string fileIn, int i1, int i2, int i3=-1, int i4=-1, int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, int i10=-1, int i11=-1, int i12=-1, int i13=-1, int i14=-1, int i15=-1, int i16=-1, int i17=-1, int i18=-1, int i19=-1) {
      if ( !exists(displayName) && settings.verbose >= 1) {
        cout<<"(VinciaRoot): ERROR! display "<<displayName
            <<" does not exist. Attempt to add histogram file failed."<<endl;
        return;
      }
      displays[displayName]->addHisFile(fileIn, i1, i2, i3, i4, i5, i6, i7, i8, i9, i10, i11, i12, i13, i14, i15, i16, i17, i18, i19);
    }    

    // Initialise a TH1D* from a file containing a saved run
    void initTH1Dptr(TH1D* hisPtrIn, string fileIn, bool hasSyst=false) {
      vector<int> format;
      format.push_back(10); format.push_back(20); format.push_back(30);
      format.push_back(40); format.push_back(51); format.push_back(52);
      if (hasSyst) { format.push_back(61); format.push_back(62); }
      VRfile initFile;      
      initFile.init();
      initFile.setFormat(format);
      initFile.readFile(fileIn);
      if (initFile.size() <= 1) return;
      for (int i=0; i<initFile.size(); ++i) {
	// Note: could do check that x match here
	if (i != 0 && i != initFile.size() -1) {
	  double xHist = hisPtrIn->GetBinCenter(i);
	  double xSave = initFile.getXC(i);
	  if (abs(xHist-xSave) > 1e-3) {
	    cout<<"VinciaRoot::initTH1Dptr(): Mismatched histogram bins"<<endl;
	    return;
	  }
	}
	hisPtrIn->SetBinContent(i,initFile.getY(i));
	hisPtrIn->SetBinError(i,max(initFile.getUstatHi(i),initFile.getUstatLo(i)));
      }
    }
    
    // Force rebin of displayed histogram to match that of data file
    void setRebinLikeData(string displayNameIn="all", bool rebinLikeData=true) {
      // If "all", set for all displays 
      if (displayNameIn == "all"|| displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.rebinLikeData = true;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setRebinLikeData(rebinLikeData);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") {
        settings.rebinLikeData  =true;
      }      
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setRebinLikeData(rebinLikeData);
      else cout<<"(VinciaRoot): setRebinLikeData no display named "
               <<displayNameIn<<endl;      
    }

    // Specify whether the average of a rebinned observable should be
    // "primitive" (ignoring zero values) or "advanced" (including zeroes)
    void setZeroSkipping(string displayNameIn="all", bool zeroSkippingIn=true) {
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.zeroSkipping = zeroSkippingIn;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setZeroSkipping(zeroSkippingIn);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") 
        settings.zeroSkipping = zeroSkippingIn;
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setZeroSkipping(zeroSkippingIn);
      else cout<<"(VinciaRoot): setZeroSkipping no display named "
               <<displayNameIn<<endl;      
    }
    
    // Set left top (white-on-black) label
    void setLeftLabel(string displayNameIn, string leftLabel) {
      // If "all", set for all displays 
      if ( displayNameIn == "all" || displayNameIn == "All" ||
            displayNameIn == "ALL") {        
        settings.leftLabel = leftLabel;
        if (leftLabel != "" && leftLabel != " ") 
          settings.hasLeftLabel = true;
        else 
          settings.leftLabel = leftLabel;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setLeftLabel(leftLabel);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") {
        if (leftLabel != "" && leftLabel != " ") 
          settings.hasLeftLabel = true;
        else 
          settings.hasLeftLabel = false;
        settings.leftLabel = leftLabel;
      }      
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setLeftLabel(leftLabel);
      else cout<<"(VinciaRoot): setLeftLabel no display named "<<displayNameIn
               <<endl;      
    }

    // Set right top (white-on-black) label
    void setRightLabel(string displayNameIn, string rightLabel) {
      // If "all", set for all displays 
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        if (rightLabel != "" && rightLabel != " ") 
          settings.hasRightLabel = true;
        else 
          settings.hasRightLabel = false;          
        settings.rightLabel = rightLabel;
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setRightLabel(rightLabel);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") {
        if (rightLabel != "" && rightLabel != " ") 
          settings.hasRightLabel = true;
        else 
          settings.hasRightLabel = false;
        settings.rightLabel = rightLabel;
      }      
      // Else set for a single (named) display
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setRightLabel(rightLabel);
      else cout<<"(VinciaRoot): setRightLabel no display named "<<displayNameIn
               <<endl;      
    }
    
    // Add objects
    void addTLine(string displayName, TLine* linePtr) {
      if ( exists(displayName) ) displays[displayName]->addTLine(linePtr);
    }

    // Update (and save) display(s) if settings function(s) tell us to
    bool update(int iEvent=-1) {
      if (iEvent < 0) ++nUpdateCallsSav;
      else nUpdateCallsSav = iEvent;
      bool doUpdate= settings.doUpdate(nUpdateCallsSav);
      bool doSave  = settings.doSave(nUpdateCallsSav);
      if (doUpdate) updateDisplay("all");
      if (doSave) {
	if (settings.saveDisplays) saveDisplay("all"); 
	if (settings.saveHistograms) saveHistogram("all");
      }
      return doUpdate;
    }     
    // Force update of single display or all displays
    // (Return false if requested display not found)
    bool updateDisplay(string displayNameIn="all") {
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) it->second->update();        
	return true;
      } else if ( exists(displayNameIn) ) {
        displays[displayNameIn]->update();        
	return true;
      }
      else return false;
    }
    // Query when next update is scheduled
    int nextUpdate(int offset=0) {
      return settings.nextUpdate(offset);
    }
    
    // Force save specific display, optionally with forced name and file type
    void saveDisplay(string displayNameIn, 
                     string baseName="none",		     
		     string fileType="default") {
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) it->second->saveDisplay(baseName,fileType);
      } else if ( exists(displayNameIn) ) {
        displays[displayNameIn]->saveDisplay(baseName,fileType);     
      } else cout<<"(VinciaRoot): saveDisplay no display named "<<displayNameIn
		 <<endl;
    }
    
    // Query when next save is scheduled
    int nextSave(int offset=0) {
      return settings.nextSave(offset);
    }

    // Save Main histograms (with uncertainties)
    void saveHistogram(string displayNameIn,
		       string baseName="none",
		       string fileType="dat") {
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) {
	  it->second->saveHistogram(baseName,  fileType);
	}
      } else if ( exists(displayNameIn) ) {
        displays[displayNameIn]->saveHistogram(baseName,fileType);     
      } else cout<<"(VinciaRoot): saveHistogram no display named "<<displayNameIn
		 <<endl;
    }

    // Return pointer to an entire display (if desired)
    VRdisplay* getDisplay(string displayName) { 
      if (exists(displayName)) return displays[displayName];
      else {
        cout<<"(VinciaRoot): ERROR! attempt to access non-existing display "
            <<displayName<<" - retuning NULL"<<endl;
        return NULL;
      }
    }

    // Print chi2 values for all displays 
    void printChi2() {
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) it->second->printChi2();
    }
    
    // Set global sigma multiplier 
    void setSigmaMC(string displayNameIn = "all", double sigmaStatIn=1.0) {
      if (displayNameIn == "all" || displayNameIn == "All" ||
          displayNameIn == "ALL") {
        settings.sigmaStat = sigmaStatIn;
        // Propagate change to all displays
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setSigmaMC(sigmaStatIn);
      }
      // If "new", do not modify existing displays, only change settings
      else if (displayNameIn == "new" || displayNameIn == "New" ||
               displayNameIn == "NEW") 
        settings.sigmaStat = sigmaStatIn;
      else if (exists(displayNameIn)) 
        displays[displayNameIn]->setSigmaMC(sigmaStatIn);
      else cout<<"(VinciaRoot): setSigmaMC no display named "
               <<displayNameIn<<endl;
    }

  private:

    // vector of displays
    map<string,VRdisplay*> displays;

    // Global settings
    VRsettings settings;
    Pythia8::Settings* settingsPtr;

    // Number of calls to update
    double nUpdateCallsSav;
    
  };

  //***************************************************************
  // class UncertaintyBand: 
  // Extremely simple auxiliary class for storing names and indices 
  // of uncertainty bands
  //***************************************************************

  class UncertaintyBand {
    
  public: 
    
    // Constructor
    UncertaintyBand(string nameIn=" ") {
      name = nameIn;
      indices.resize(0);
    };
    
    // Name
    string name;
    // Indices
    vector<int> indices;
    
  };
 
}

#endif
